# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import List

from gravitino.api.expressions.expression import Expression
from gravitino.api.expressions.sorts.null_ordering import NullOrdering
from gravitino.api.expressions.sorts.sort_direction import SortDirection
from gravitino.api.expressions.sorts.sort_order import SortOrder


class SortImpl(SortOrder):

    def __init__(
        self,
        expression: Expression,
        direction: SortDirection,
        null_ordering: NullOrdering,
    ):
        """Initialize the SortImpl object."""
        self._expression = expression
        self._direction = direction
        self._null_ordering = null_ordering

    def expression(self) -> Expression:
        return self._expression

    def direction(self) -> SortDirection:
        return self._direction

    def null_ordering(self) -> NullOrdering:
        return self._null_ordering

    def __eq__(self, other: object) -> bool:
        """Check if two SortImpl instances are equal."""
        if not isinstance(other, SortImpl):
            return False
        return (
            self.expression() == other.expression()
            and self.direction() == other.direction()
            and self.null_ordering() == other.null_ordering()
        )

    def __hash__(self) -> int:
        """Generate a hash for a SortImpl instance."""
        return hash((self.expression(), self.direction(), self.null_ordering()))

    def __str__(self) -> str:
        """Provide a string representation of the SortImpl object."""
        return (
            f"SortImpl(expression={self._expression}, "
            f"direction={self._direction}, null_ordering={self._null_ordering})"
        )


class SortOrders:
    """Helper methods to create SortOrders to pass into Apache Gravitino."""

    # NONE is used to indicate that there is no sort order
    NONE: List[SortOrder] = []

    @staticmethod
    def ascending(expression: Expression) -> SortImpl:
        """Creates a sort order with ascending direction and nulls first."""
        return SortOrders.of(expression, SortDirection.ASCENDING)

    @staticmethod
    def descending(expression: Expression) -> SortImpl:
        """Creates a sort order with descending direction and nulls last."""
        return SortOrders.of(expression, SortDirection.DESCENDING)

    @staticmethod
    def of(
        expression: Expression,
        direction: SortDirection,
        null_ordering: NullOrdering = None,
    ) -> SortImpl:
        """Creates a sort order with the given direction and optionally specified null ordering."""
        if null_ordering is None:
            null_ordering = direction.default_null_ordering()
        return SortImpl(expression, direction, null_ordering)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field

from dataclasses_json import config

from gravitino.rest.rest_message import RESTRequest


@dataclass
class MetalakeSetRequest(RESTRequest):
    """Represents a request to set a metalake in use status."""

    _in_use: bool = field(metadata=config(field_name="inUse"))

    def __init__(self, in_use: bool):
        self._in_use = in_use

    def validate(self):
        """Validates the fields of the request.

        Raises:
            IllegalArgumentException if in_use is not set.
        """
        if self._in_use is None:
            raise ValueError('"in_use" field is required and cannot be empty')



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import abstractmethod
from typing import Optional, Dict

from gravitino.api.auditable import Auditable


class Schema(Auditable):
    """
    An interface representing a schema in the Catalog. A Schema is a
    basic container of relational objects, like tables, views, etc. A Schema can be self-nested,
    which means it can be schema1.schema2.table.

    This defines the basic properties of a schema. A catalog implementation with SupportsSchemas
    should implement this interface.
    """

    @abstractmethod
    def name(self) -> str:
        """Returns the name of the Schema."""
        pass

    def comment(self) -> Optional[str]:
        """Returns the comment of the Schema. None is returned if the comment is not set."""
        return None

    def properties(self) -> Dict[str, str]:
        """Returns the properties of the Schema. An empty dictionary is returned if no properties are set."""
        return {}



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import abstractmethod
from typing import Optional

from gravitino.utils.http_client import HTTPClient
from gravitino.auth.auth_data_provider import AuthDataProvider
from gravitino.auth.auth_constants import AuthConstants


class OAuth2TokenProvider(AuthDataProvider):
    """OAuth2TokenProvider will request the access token from the authorization server and then provide
    the access token for every request.
    """

    # The HTTP client used to request the access token from the authorization server.
    _client: HTTPClient

    def __init__(self, uri: str):
        self._client = HTTPClient(uri)

    def has_token_data(self) -> bool:
        """Judge whether AuthDataProvider can provide token data.

        Returns:
            true if the AuthDataProvider can provide token data otherwise false.
        """
        return True

    def get_token_data(self) -> Optional[bytes]:
        """Acquire the data of token for authentication. The client will set the token data as HTTP header
        Authorization directly. So the return value should ensure token data contain the token header
        (eg: Bearer, Basic) if necessary.

        Returns:
            the token data is used for authentication.
        """
        access_token = self._get_access_token()

        if access_token is None:
            return None

        return (AuthConstants.AUTHORIZATION_BEARER_HEADER + access_token).encode(
            "utf-8"
        )

    def close(self):
        """Closes the OAuth2TokenProvider and releases any underlying resources."""
        if self._client is not None:
            self._client.close()

    @abstractmethod
    def _get_access_token(self) -> Optional[str]:
        """Get the access token from the authorization server."""

    @abstractmethod
    def validate(self):
        """Validate the OAuth2TokenProvider"""



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import abstractmethod
from typing import List, TypeVar, Generic

from gravitino.api.expressions.expression import Expression
from gravitino.api.types.type import Type

T = TypeVar("T")


class Literal(Generic[T], Expression):
    """
    Represents a constant literal value in the public expression API.
    """

    @abstractmethod
    def value(self) -> T:
        """The literal value."""
        raise NotImplementedError("Subclasses must implement the `value` method.")

    @abstractmethod
    def data_type(self) -> Type:
        """The data type of the literal."""
        raise NotImplementedError("Subclasses must implement the `data_type` method.")

    def children(self) -> List[Expression]:
        return Expression.EMPTY_EXPRESSION



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import base64
import os

from gravitino.auth.auth_constants import AuthConstants
from gravitino.auth.auth_data_provider import AuthDataProvider


class SimpleAuthProvider(AuthDataProvider):
    """SimpleAuthProvider will use the environment variable `GRAVITINO_USER` or
    the user of the system to generate a basic token for every request.

    """

    _token: bytes

    def __init__(self):
        gravitino_user = os.environ.get("GRAVITINO_USER")
        if gravitino_user is None or len(gravitino_user) == 0:
            gravitino_user = os.environ.get("user.name")

        if gravitino_user is None or len(gravitino_user) == 0:
            gravitino_user = "anonymous"

        user_information = f"{gravitino_user}:dummy"
        self._token = (
            AuthConstants.AUTHORIZATION_BASIC_HEADER
            + base64.b64encode(user_information.encode("utf-8")).decode("utf-8")
        ).encode("utf-8")

    def has_token_data(self) -> bool:
        return True

    def get_token_data(self) -> bytes:
        return self._token

    def close(self):
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC
from dataclasses import dataclass, field

from dataclasses_json import config


class SchemaChange(ABC):
    """NamespaceChange class to set the property and value pairs for the namespace."""

    @staticmethod
    def set_property(schema_property: str, value: str):
        """SchemaChange class to set the property and value pairs for the schema.

        Args:
            schema_property: The property name to set.
            value: The value to set the property to.

        Returns:
             The SchemaChange object.
        """
        return SchemaChange.SetProperty(schema_property, value)

    @staticmethod
    def remove_property(schema_property: str):
        """SchemaChange class to remove a property from the schema.

        Args:
            schema_property: The property name to remove.

        Returns:
            The SchemaChange object.
        """
        return SchemaChange.RemoveProperty(schema_property)

    @dataclass
    class SetProperty:
        """SchemaChange class to set the property and value pairs for the schema."""

        _property: str = field(metadata=config(field_name="property"))
        _value: str = field(metadata=config(field_name="value"))

        def property(self):
            """Retrieves the name of the property to be set.

            Returns:
                 The name of the property.
            """
            return self._property

        def value(self):
            """Retrieves the value of the property to be set.

            Returns:
                 The value of the property.
            """
            return self._value

        def __eq__(self, other):
            """Compares this SetProperty instance with another object for equality.
            Two instances are considered equal if they have the same property and value.

            Args:
                 other: The object to compare with this instance.

            Returns:
                 true if the given object represents the same property setting; false otherwise.
            """
            if not isinstance(other, SchemaChange.SetProperty):
                return False
            return self._property == other.property() and self._value == other.value()

        def __hash__(self):
            """Generates a hash code for this SetProperty instance.
            The hash code is based on both the property name and its value.

             Returns:
                  A hash code value for this property setting.
            """
            return hash((self._property, self._value))

        def __str__(self):
            """Provides a string representation of the SetProperty instance.
            This string format includes the class name followed by the property name and its value.

            Returns:
                 A string summary of the property setting.
            """
            return f"SETPROPERTY {self._property} {self._value}"

    @dataclass
    class RemoveProperty:
        """SchemaChange class to remove a property from the schema."""

        _property: str = field(metadata=config(field_name="property"))

        def property(self):
            """Retrieves the name of the property to be removed.

            Returns:
                 The name of the property for removal.
            """
            return self._property

        def __eq__(self, other):
            """Compares this RemoveProperty instance with another object for equality.
            Two instances are considered equal if they target the same property for removal.

            Args:
                other: The object to compare with this instance.

            Returns:
                 true if the given object represents the same property removal; false otherwise.
            """
            if not isinstance(other, SchemaChange.RemoveProperty):
                return False
            return self._property == other.property()

        def __hash__(self):
            """Generates a hash code for this RemoveProperty instance.
            This hash code is based on the property name that is to be removed.

            Returns:
                 A hash code value for this property removal operation.
            """
            return hash(self._property)

        def __str__(self):
            """Provides a string representation of the RemoveProperty instance.
            This string format includes the class name followed by the property name to be removed.

            Returns:
                 A string summary of the property removal operation.
            """
            return f"REMOVEPROPERTY {self._property}"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import List, Optional
from dataclasses import dataclass, field
from dataclasses_json import config

from gravitino.dto.responses.base_response import BaseResponse
from gravitino.constants.error import ErrorConstants, EXCEPTION_MAPPING
from gravitino.exceptions.base import BadRequestException, UnknownError


@dataclass
class ErrorResponse(BaseResponse):
    """Represents an error response."""

    _type: str = field(metadata=config(field_name="type"))
    _message: str = field(metadata=config(field_name="message"))
    _stack: Optional[List[str]] = field(metadata=config(field_name="stack"))

    def type(self):
        return self._type

    def message(self):
        return self._message

    def stack(self):
        return self._stack

    def validate(self):
        super().validate()

        if self._type is None or not self._type.strip():
            raise BadRequestException("Type cannot be None or empty")

        if self._message is None or not self._message.strip():
            raise BadRequestException("Message cannot be None or empty")

    def __repr__(self) -> str:
        return (
            f"ErrorResponse(code={self._code}, type={self._type}, message={self._message})"
            + ("\n\t" + "\n\t".join(self._stack) if self._stack is not None else "")
        )

    def format_error_message(self) -> str:
        return (
            f"{self._message}\n" + "\n".join(self._stack)
            if self._stack is not None
            else ""
        )

    @classmethod
    def generate_error_response(
        cls, exception: Exception, message: str
    ) -> "ErrorResponse":
        for exception_type, error_code in EXCEPTION_MAPPING.items():
            if issubclass(exception, exception_type):
                return cls(error_code, exception.__name__, message, None)

        return cls(
            ErrorConstants.UNKNOWN_ERROR_CODE, UnknownError.__name__, message, None
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from dataclasses_json import config


@dataclass
class VersionDTO:
    """Represents a Version Data Transfer Object (DTO)."""

    _version: str = field(metadata=config(field_name="version"))
    """The version of the software."""

    _compile_date: str = field(metadata=config(field_name="compileDate"))
    """The date the software was compiled."""

    _git_commit: str = field(metadata=config(field_name="gitCommit"))
    """The git commit of the software."""

    def version(self) -> str:
        return self._version

    def compile_date(self) -> str:
        return self._compile_date

    def git_commit(self) -> str:
        return self._git_commit



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod


class AuthDataProvider(ABC):
    """The provider of authentication data"""

    @abstractmethod
    def has_token_data(self) -> bool:
        """Judge whether AuthDataProvider can provide token data.

        Returns:
            true if the AuthDataProvider can provide token data otherwise false.
        """
        pass

    @abstractmethod
    def get_token_data(self) -> bytes:
        """Acquire the data of token for authentication. The client will set the token data as HTTP header
        Authorization directly. So the return value should ensure token data contain the token header
        (eg: Bearer, Basic) if necessary.

        Returns:
            the token data is used for authentication.
        """

    @abstractmethod
    def close(self):
        """Close the resource in the provider."""



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from enum import Enum


class MetadataObject(ABC):
    """The MetadataObject is the basic unit of the Gravitino system. It
    represents the metadata object in the Apache Gravitino system. The object
    can be a metalake, catalog, schema, table, topic, etc.
    """

    class Type(Enum):
        """The type of object in the Gravitino system. Every type will map one
        kind of the entity of the underlying system."""

        CATALOG = "catalog"
        """"Metadata Type for catalog."""

        FILESET = "fileset"
        """Metadata Type for Fileset System (including HDFS, S3, etc.), like path/to/file"""

    @abstractmethod
    def type(self) -> Type:
        """
        The type of the object.

        Returns:
            The type of the object.
        """
        pass

    @abstractmethod
    def name(self) -> str:
        """
        The name of the object.

        Returns:
            The name of the object.
        """
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from gravitino.exceptions.base import UnauthorizedException, BadRequestException
from gravitino.dto.responses.oauth2_error_response import OAuth2ErrorResponse
from gravitino.exceptions.handlers.rest_error_handler import RestErrorHandler

INVALID_CLIENT_ERROR = "invalid_client"
INVALID_REQUEST_ERROR = "invalid_request"
INVALID_GRANT_ERROR = "invalid_grant"
UNAUTHORIZED_CLIENT_ERROR = "unauthorized_client"
UNSUPPORTED_GRANT_TYPE_ERROR = "unsupported_grant_type"
INVALID_SCOPE_ERROR = "invalid_scope"


class OAuthErrorHandler(RestErrorHandler):

    def handle(self, error_response: OAuth2ErrorResponse):

        error_message = error_response.message()
        exception_type = error_response.type()

        if exception_type == INVALID_CLIENT_ERROR:
            raise UnauthorizedException(
                f"Not authorized: {exception_type}: {error_message}"
            )

        if exception_type in [
            INVALID_REQUEST_ERROR,
            INVALID_GRANT_ERROR,
            UNAUTHORIZED_CLIENT_ERROR,
            UNSUPPORTED_GRANT_TYPE_ERROR,
            INVALID_SCOPE_ERROR,
        ]:
            raise BadRequestException(
                f"Malformed request: {exception_type}: {error_message}"
            )

        super().handle(error_response)


OAUTH_ERROR_HANDLER = OAuthErrorHandler()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field

from dataclasses_json import DataClassJsonMixin, config

from gravitino.dto.responses.base_response import BaseResponse
from gravitino.dto.schema_dto import SchemaDTO
from gravitino.exceptions.base import IllegalArgumentException


@dataclass
class SchemaResponse(BaseResponse, DataClassJsonMixin):
    """Represents a response for a schema."""

    _schema: SchemaDTO = field(metadata=config(field_name="schema"))

    # TODO
    # pylint: disable=arguments-differ
    def schema(self) -> SchemaDTO:
        return self._schema

    def validate(self):
        """Validates the response data.

        Raises:
            IllegalArgumentException if catalog identifiers are not set.
        """
        super().validate()

        if self._schema is None:
            raise IllegalArgumentException("Schema must be non-null")

        if self._schema.name() is None:
            raise IllegalArgumentException("Schema 'name' must not be null and empty")

        if self._schema.audit_info() is None:
            raise IllegalArgumentException("Schema 'audit' must not be null")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field

from dataclasses_json import config

from gravitino.dto.fileset_dto import FilesetDTO
from gravitino.dto.responses.base_response import BaseResponse
from gravitino.exceptions.base import IllegalArgumentException


@dataclass
class FilesetResponse(BaseResponse):
    """Response for fileset creation."""

    _fileset: FilesetDTO = field(metadata=config(field_name="fileset"))

    def fileset(self) -> FilesetDTO:
        return self._fileset

    def validate(self):
        """Validates the response data.

        Raises:
            IllegalArgumentException if catalog identifiers are not set.
        """
        super().validate()
        if self._fileset is None:
            raise IllegalArgumentException("fileset must not be null")
        if not self._fileset.name():
            raise IllegalArgumentException("fileset 'name' must not be null and empty")
        if not self._fileset.storage_location():
            raise IllegalArgumentException(
                "fileset 'storageLocation' must not be null and empty"
            )
        if self._fileset.type() is None:
            raise IllegalArgumentException("fileset 'type' must not be null and empty")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import List, Dict

from gravitino.api.catalog import Catalog
from gravitino.api.catalog_change import CatalogChange
from gravitino.auth.auth_data_provider import AuthDataProvider
from gravitino.client.gravitino_client_base import GravitinoClientBase
from gravitino.client.gravitino_metalake import GravitinoMetalake


class GravitinoClient(GravitinoClientBase):
    """Gravitino Client for a user to interact with the Gravitino API, allowing the client to list,
    load, create, and alter Catalog.

    It uses an underlying {@link RESTClient} to send HTTP requests and receive responses from the API.
    """

    _metalake: GravitinoMetalake

    def __init__(
        self,
        uri: str,
        metalake_name: str,
        check_version: bool = True,
        auth_data_provider: AuthDataProvider = None,
    ):
        """Constructs a new GravitinoClient with the given URI, authenticator and AuthDataProvider.

        Args:
            uri: The base URI for the Gravitino API.
            metalake_name: The specified metalake name.
            auth_data_provider: The provider of the data which is used for authentication.

        Raises:
            NoSuchMetalakeException if the metalake with specified name does not exist.
        """
        super().__init__(uri, check_version, auth_data_provider)
        self.check_metalake_name(metalake_name)
        self._metalake = super().load_metalake(metalake_name)

    def get_metalake(self) -> GravitinoMetalake:
        """Get the current metalake object

        Raises:
            NoSuchMetalakeException if the metalake with specified name does not exist.

        Returns:
            the GravitinoMetalake object
        """
        return self._metalake

    def list_catalogs(self) -> List[str]:
        return self.get_metalake().list_catalogs()

    def list_catalogs_info(self) -> List[Catalog]:
        return self.get_metalake().list_catalogs_info()

    def load_catalog(self, name: str) -> Catalog:
        return self.get_metalake().load_catalog(name)

    def create_catalog(
        self,
        name: str,
        catalog_type: Catalog.Type,
        provider: str,
        comment: str,
        properties: Dict[str, str],
    ) -> Catalog:
        return self.get_metalake().create_catalog(
            name, catalog_type, provider, comment, properties
        )

    def alter_catalog(self, name: str, *changes: CatalogChange):
        return self.get_metalake().alter_catalog(name, *changes)

    def drop_catalog(self, name: str, force: bool = False) -> bool:
        return self.get_metalake().drop_catalog(name, force)

    def enable_catalog(self, name: str):
        return self.get_metalake().enable_catalog(name)

    def disable_catalog(self, name: str):
        return self.get_metalake().disable_catalog(name)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from dataclasses import dataclass, field
from typing import Optional, Dict

from dataclasses_json import DataClassJsonMixin, config

from gravitino.api.model import Model
from gravitino.dto.audit_dto import AuditDTO


@dataclass
class ModelDTO(Model, DataClassJsonMixin):
    """Represents a Model DTO (Data Transfer Object)."""

    _name: str = field(metadata=config(field_name="name"))
    _comment: Optional[str] = field(metadata=config(field_name="comment"))
    _properties: Optional[Dict[str, str]] = field(
        metadata=config(field_name="properties")
    )
    _latest_version: int = field(metadata=config(field_name="latestVersion"))
    _audit: AuditDTO = field(default=None, metadata=config(field_name="audit"))

    def name(self) -> str:
        return self._name

    def comment(self) -> Optional[str]:
        return self._comment

    def properties(self) -> Optional[Dict[str, str]]:
        return self._properties

    def latest_version(self) -> int:
        return self._latest_version

    def audit_info(self) -> AuditDTO:
        return self._audit



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from typing import Optional, Dict

from dataclasses_json import DataClassJsonMixin, config

from gravitino.api.audit import Audit
from gravitino.dto.audit_dto import AuditDTO
from gravitino.api.metalake import Metalake


@dataclass
class MetalakeDTO(Metalake, DataClassJsonMixin):
    """Represents a Metalake Data Transfer Object (DTO) that implements the Metalake interface."""

    _name: str = field(metadata=config(field_name="name"))
    """The name of the Metalake DTO."""

    _comment: Optional[str] = field(metadata=config(field_name="comment"))
    """The comment of the Metalake DTO."""

    _properties: Optional[Dict[str, str]] = field(
        metadata=config(field_name="properties")
    )
    """The properties of the Metalake DTO."""

    _audit: Optional[AuditDTO] = field(metadata=config(field_name="audit"))
    """The audit information of the Metalake DTO."""

    def name(self) -> str:
        return self._name

    def comment(self) -> str:
        return self._comment

    def properties(self) -> Dict[str, str]:
        return self._properties

    def audit_info(self) -> Audit:
        return self._audit

    def equals(self, other):
        if self == other:
            return True
        if not isinstance(other, MetalakeDTO):
            return False
        return (
            self.name() == other.name()
            and self.comment() == other.comment()
            and self.property_equal(self.properties(), other.properties())
            and self.audit_info() == other.audit_info()
        )

    def property_equal(self, p1, p2):
        if p1 is None and p2 is None:
            return True
        if p1 is not None and not p1 and p2 is None:
            return True
        if p2 is not None and not p2 and p1 is None:
            return True
        return p1 == p2



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from enum import Enum


class NullOrdering(Enum):
    """A null order used in sorting expressions."""

    NULLS_FIRST: str = "nulls_first"
    """Nulls appear before non-nulls. For ascending order, this means nulls appear at the beginning."""

    NULLS_LAST: str = "nulls_last"
    """Nulls appear after non-nulls. For ascending order, this means nulls appear at the end."""

    def __str__(self) -> str:
        if self == NullOrdering.NULLS_FIRST:
            return "nulls_first"
        if self == NullOrdering.NULLS_LAST:
            return "nulls_last"

        raise ValueError(f"Unexpected null order: {self}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from enum import Enum


class FilesetDataOperation(Enum):
    """An enum class containing fileset data operations that supported."""

    CREATE = "CREATE"
    """Creates a new file.
    """

    OPEN = "OPEN"
    """Opens a file.
    """

    OPEN_AND_WRITE = "OPEN_AND_WRITE"
    """Opens a file and writes to it.
    """

    OPEN_AND_APPEND = "OPEN_AND_APPEND"
    """Opens a file and appends to it.
    """

    APPEND = "APPEND"
    """Appends some content into a file.
    """

    RENAME = "RENAME"
    """Renames a file or a directory.
    """

    DELETE = "DELETE"
    """Deletes a file or a directory.
    """

    GET_FILE_STATUS = "GET_FILE_STATUS"
    """Gets a file status from a file or a directory.
    """

    LIST_STATUS = "LIST_STATUS"
    """Lists file statuses under a directory.
    """

    MKDIRS = "MKDIRS"
    """Creates a directory.
    """

    EXISTS = "EXISTS"
    """Checks if a file or a directory exists.
    """

    CREATED_TIME = "CREATED_TIME"
    """Gets the created time of a file.
    """

    MODIFIED_TIME = "MODIFIED_TIME"
    """Gets the modified time of a file.
    """

    COPY_FILE = "COPY_FILE"
    """Copies a file.
    """

    CAT_FILE = "CAT_FILE"
    """Gets the content of a file.
    """

    GET_FILE = "GET_FILE"
    """Copies a remote file to local.
    """

    UNKNOWN = "UNKNOWN"
    """Unknown data operation.
    """



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
# pylint: disable=C0302
from __future__ import annotations
from typing import List
from .type import (
    Type,
    Name,
    PrimitiveType,
    IntegralType,
    FractionType,
    DateTimeType,
    IntervalType,
    ComplexType,
)


class Types:
    """The helper class for Type. It contains all built-in types and provides utility methods."""

    class NullType(Type):
        """The data type representing `NULL` values."""

        _instance: Types.NullType = None

        def __new__(cls):
            if cls._instance is None:
                cls._instance = super(Types.NullType, cls).__new__(cls)
            return cls._instance

        @classmethod
        def get(cls) -> Types.NullType:
            return cls()

        def name(self) -> Name:
            return Name.NULL

        def simple_string(self) -> str:
            return "null"

    class BooleanType(PrimitiveType):
        """The boolean type in Gravitino."""

        _instance: Types.BooleanType = None

        def __new__(cls):
            if cls._instance is None:
                cls._instance = super(Types.BooleanType, cls).__new__(cls)
            return cls._instance

        @classmethod
        def get(cls) -> Types.BooleanType:
            return cls()

        def name(self) -> Name:
            return Name.BOOLEAN

        def simple_string(self) -> str:
            return "boolean"

    class ByteType(IntegralType):
        """The byte type in Gravitino."""

        _instance: Types.ByteType = None
        _unsigned_instance: Types.ByteType = None

        def __new__(cls, signed: bool = True):
            if signed:
                if cls._instance is None:
                    cls._instance = super(Types.ByteType, cls).__new__(cls)
                    cls._instance.__init__(signed)
                return cls._instance
            if cls._unsigned_instance is None:
                cls._unsigned_instance = super(Types.ByteType, cls).__new__(cls)
                cls._unsigned_instance.__init__(signed)
            return cls._unsigned_instance

        @classmethod
        def get(cls) -> Types.ByteType:
            return cls(True)

        @classmethod
        def unsigned(cls) -> Types.ByteType:
            return cls(False)

        def name(self) -> Name:
            return Name.BYTE

        def simple_string(self) -> str:
            return "byte" if self.signed() else "byte unsigned"

    class ShortType(IntegralType):
        _instance: Types.ShortType = None
        _unsigned_instance: Types.ShortType = None

        def __new__(cls, signed=True):
            if signed:
                if cls._instance is None:
                    cls._instance = super(Types.ShortType, cls).__new__(cls)
                    cls._instance.__init__(signed)
                return cls._instance
            if cls._unsigned_instance is None:
                cls._unsigned_instance = super(Types.ShortType, cls).__new__(cls)
                cls._unsigned_instance.__init__(signed)
            return cls._unsigned_instance

        @classmethod
        def get(cls) -> Types.ShortType:
            return cls(True)

        @classmethod
        def unsigned(cls) -> Types.ShortType:
            return cls(False)

        def name(self) -> Name:
            return Name.SHORT

        def simple_string(self) -> str:
            return "short" if self.signed() else "short unsigned"

    class IntegerType(IntegralType):
        _instance: Types.IntegerType = None
        _unsigned_instance: Types.IntegerType = None

        def __new__(cls, signed=True):
            if signed:
                if cls._instance is None:
                    cls._instance = super(Types.IntegerType, cls).__new__(cls)
                    cls._instance.__init__(signed)
                return cls._instance
            if cls._unsigned_instance is None:
                cls._unsigned_instance = super(Types.IntegerType, cls).__new__(cls)
                cls._unsigned_instance.__init__(signed)
            return cls._unsigned_instance

        @classmethod
        def get(cls) -> Types.IntegerType:
            return cls(True)

        @classmethod
        def unsigned(cls):
            return cls(False)

        def name(self) -> Name:
            return Name.INTEGER

        def simple_string(self) -> str:
            return "integer" if self.signed() else "integer unsigned"

    class LongType(IntegralType):
        _instance: Types.LongType = None
        _unsigned_instance: Types.LongType = None

        def __new__(cls, signed=True):
            if signed:
                if cls._instance is None:
                    cls._instance = super(Types.LongType, cls).__new__(cls)
                    cls._instance.__init__(signed)
                return cls._instance
            if cls._unsigned_instance is None:
                cls._unsigned_instance = super(Types.LongType, cls).__new__(cls)
                cls._unsigned_instance.__init__(signed)
            return cls._unsigned_instance

        @classmethod
        def get(cls) -> Types.LongType:
            return cls(True)

        @classmethod
        def unsigned(cls):
            return cls(False)

        def name(self) -> Name:
            return Name.LONG

        def simple_string(self) -> str:
            return "long" if self.signed() else "long unsigned"

    class FloatType(FractionType):
        _instance: Types.FloatType = None

        def __new__(cls):
            if cls._instance is None:
                cls._instance = super(Types.FloatType, cls).__new__(cls)
                cls._instance.__init__()
            return cls._instance

        @classmethod
        def get(cls) -> Types.FloatType:
            return cls()

        def name(self) -> Name:
            return Name.FLOAT

        def simple_string(self) -> str:
            return "float"

    class DoubleType(FractionType):
        _instance: Types.DoubleType = None

        def __new__(cls):
            if cls._instance is None:
                cls._instance = super(Types.DoubleType, cls).__new__(cls)
                cls._instance.__init__()
            return cls._instance

        @classmethod
        def get(cls) -> Types.DoubleType:
            return cls()

        def name(self) -> Name:
            return Name.DOUBLE

        def simple_string(self) -> str:
            return "double"

    class DecimalType(FractionType):
        """The decimal type in Gravitino."""

        MAX_PRECISION = 38
        _precision: int
        _scale: int

        def __init__(self, precision: int, scale: int):
            """
            Args:
                precision: The precision of the decimal type.
                scale: The scale of the decimal type.
            """
            super().__init__()
            self.check_precision_scale(precision, scale)
            self._precision = precision
            self._scale = scale

        @staticmethod
        def check_precision_scale(precision: int, scale: int):
            """
            Ensures the precision and scale values are within valid range.

            Args:
                precision: The precision of the decimal.
                scale: The scale of the decimal.
            """
            if not 1 <= precision <= Types.DecimalType.MAX_PRECISION:
                raise ValueError(
                    f"Decimal precision must be in range [1, 38]: {precision}"
                )
            if not 0 <= scale <= precision:
                raise ValueError(
                    f"Decimal scale must be in range [0, precision ({precision})]: {scale}"
                )

        @classmethod
        def of(cls, precision: int, scale: int) -> Types.DecimalType:
            return cls(precision, scale)

        def name(self) -> Name:
            return Name.DECIMAL

        def precision(self) -> int:
            return self._precision

        def scale(self) -> int:
            return self._scale

        def simple_string(self) -> str:
            return f"decimal({self._precision},{self._scale})"

        def __eq__(self, other):
            """
            Compares two DecimalType objects for equality.

            Args:
                other: The other DecimalType to compare with.

            Returns:
                True if both objects have the same precision and scale, False otherwise.
            """
            if not isinstance(other, Types.DecimalType):
                return False
            return self._precision == other._precision and self._scale == other._scale

        def __hash__(self):
            return hash((self._precision, self._scale))

    class DateType(DateTimeType):
        """The date time type in Gravitino."""

        _instance: Types.DateType = None

        def __new__(cls):
            if cls._instance is None:
                cls._instance = super(Types.DateType, cls).__new__(cls)
                cls._instance.__init__()
            return cls._instance

        @classmethod
        def get(cls) -> Types.DateType:
            return cls()

        def name(self) -> Name:
            return Name.DATE

        def simple_string(self) -> str:
            return "date"

    class TimeType(DateTimeType):
        _instance: Types.TimeType = None

        def __new__(cls):
            if cls._instance is None:
                cls._instance = super(Types.TimeType, cls).__new__(cls)
                cls._instance.__init__()
            return cls._instance

        @classmethod
        def get(cls) -> Types.TimeType:
            return cls()

        def name(self) -> Name:
            return Name.TIME

        def simple_string(self) -> str:
            return "time"

    class TimestampType(DateTimeType):
        _instance_with_tz: Types.TimestampType = None
        _instance_without_tz: Types.TimestampType = None
        _with_time_zone: bool

        def __new__(cls, with_time_zone: bool):
            if with_time_zone:
                if cls._instance_with_tz is None:
                    cls._instance_with_tz = super(Types.TimestampType, cls).__new__(cls)
                    cls._instance_with_tz.__init__(with_time_zone)
                return cls._instance_with_tz
            if cls._instance_without_tz is None:
                cls._instance_without_tz = super(Types.TimestampType, cls).__new__(cls)
                cls._instance_without_tz.__init__(with_time_zone)
            return cls._instance_without_tz

        @classmethod
        def with_time_zone(cls) -> Types.TimestampType:
            return cls(True)

        @classmethod
        def without_time_zone(cls) -> Types.TimestampType:
            return cls(False)

        def __init__(self, with_time_zone: bool):
            self._with_time_zone = with_time_zone
            super().__init__()

        def has_time_zone(self) -> bool:
            return self._with_time_zone

        def name(self) -> Name:
            return Name.TIMESTAMP

        def simple_string(self) -> str:
            return "timestamp_tz" if self._with_time_zone else "timestamp"

    class IntervalYearType(IntervalType):
        """The interval year type in Gravitino."""

        _instance: Types.IntervalYearType = None

        def __new__(cls):
            if cls._instance is None:
                cls._instance = super(Types.IntervalYearType, cls).__new__(cls)
                cls._instance.__init__()
            return cls._instance

        @classmethod
        def get(cls) -> Types.IntervalYearType:
            return cls()

        def name(self) -> Name:
            return Name.INTERVAL_YEAR

        def simple_string(self) -> str:
            return "interval_year"

    class IntervalDayType(IntervalType):
        """The interval day type in Gravitino."""

        _instance: Types.IntervalDayType = None

        def __new__(cls):
            if cls._instance is None:
                cls._instance = super(Types.IntervalDayType, cls).__new__(cls)
                cls._instance.__init__()
            return cls._instance

        @classmethod
        def get(cls) -> Types.IntervalDayType:
            return cls()

        def name(self) -> Name:
            return Name.INTERVAL_DAY

        def simple_string(self) -> str:
            return "interval_day"

    class StringType(PrimitiveType):
        """The string type in Gravitino, equivalent to varchar(MAX),
        which the MAX is determined by the underlying catalog."""

        _instance: Types.StringType = None

        def __new__(cls):
            if cls._instance is None:
                cls._instance = super(Types.StringType, cls).__new__(cls)
                cls._instance.__init__()
            return cls._instance

        @classmethod
        def get(cls) -> Types.StringType:
            return cls()

        def name(self) -> Name:
            return Name.STRING

        def simple_string(self) -> str:
            return "string"

    class UUIDType(PrimitiveType):
        """The uuid type in Gravitino."""

        _instance: Types.UUIDType = None

        def __new__(cls):
            if cls._instance is None:
                cls._instance = super(Types.UUIDType, cls).__new__(cls)
                cls._instance.__init__()
            return cls._instance

        @classmethod
        def get(cls) -> Types.UUIDType:
            return cls()

        def name(self) -> Name:
            return Name.UUID

        def simple_string(self) -> str:
            return "uuid"

    class FixedType(PrimitiveType):
        """Fixed-length byte array type, if you want to use variable-length
        byte array, use BinaryType instead."""

        _length: int

        def __init__(self, length: int):
            """
            Initializes the FixedType with the given length.

            Args:
                length: The length of the fixed type.
            """
            self._length = length

        @classmethod
        def of(cls, length: int) -> Types.FixedType:
            """
            Args:
                length: The length of the fixed type.

            Returns:
                 A FixedType instance with the given length.
            """
            return cls(length)

        def name(self) -> Name:
            return Name.FIXED

        def length(self) -> int:
            return self._length

        def simple_string(self) -> str:
            return f"fixed({self._length})"

        def __eq__(self, other):
            """
            Compares two FixedType objects for equality.

            Args:
                other: The other FixedType object to compare with.

            Returns:
                True if both FixedType objects have the same length, False otherwise.
            """
            if not isinstance(other, Types.FixedType):
                return False
            return self._length == other._length

        def __hash__(self):
            return hash(self._length)

    class VarCharType(PrimitiveType):
        """The varchar type in Gravitino."""

        _length: int

        def __init__(self, length: int):
            self._length = length

        @classmethod
        def of(cls, length: int) -> Types.VarCharType:
            return cls(length)

        def name(self) -> Name:
            return Name.VARCHAR

        def length(self) -> int:
            return self._length

        def simple_string(self) -> str:
            return f"varchar({self._length})"

        def __eq__(self, other):
            """
            Compares two VarCharType objects for equality.

            Args:
                other: The other VarCharType object to compare with.

            Returns:
                True if both VarCharType objects have the same length, False otherwise.
            """
            if isinstance(other, Types.VarCharType):
                return self._length == other._length
            return False

        def __hash__(self):
            return hash(self._length)

    class FixedCharType(PrimitiveType):
        """The fixed char type in Gravitino."""

        _length: int

        def __init__(self, length: int):
            self._length = length

        @classmethod
        def of(cls, length: int) -> Types.FixedCharType:
            return cls(length)

        def name(self) -> Name:
            return Name.FIXEDCHAR

        def length(self) -> int:
            return self._length

        def simple_string(self) -> str:
            return f"char({self._length})"

        def __eq__(self, other):
            if not isinstance(other, Types.FixedCharType):
                return False
            return self._length == other._length

        def __hash__(self):
            return hash(self._length)

    class BinaryType(PrimitiveType):
        _instance: Types.BinaryType = None

        def __new__(cls):
            if cls._instance is None:
                cls._instance = super(Types.BinaryType, cls).__new__(cls)
                cls._instance.__init__()
            return cls._instance

        @classmethod
        def get(cls) -> Types.BinaryType:
            return cls()

        def name(self) -> Name:
            return Name.BINARY

        def simple_string(self) -> str:
            return "binary"

    class StructType(ComplexType):
        """The struct type in Gravitino.
        Note, this type is not supported in the current version of Gravitino."""

        _fields: List[Field]

        def __init__(self, fields: List[Field]):
            if not fields or len(fields) == 0:
                raise ValueError("fields cannot be null or empty")
            self._fields = fields

        @classmethod
        def of(cls, *fields) -> Types.StructType:
            """
            Args:
                fields: The fields of the struct type.

            Returns:
                A StructType instance with the given fields.
            """
            return cls(list(fields))

        def fields(self) -> List[Field]:
            return self._fields

        def name(self) -> Name:
            return Name.STRUCT

        def simple_string(self) -> str:
            return (
                f"struct<{', '.join(field.simple_string() for field in self._fields)}>"
            )

        def __eq__(self, other):
            """
            Compares two StructType objects for equality.

            Args:
                other: The other StructType object to compare with.

            Returns:
                True if both StructType objects have the same fields, False otherwise.
            """
            if isinstance(other, Types.StructType):
                return self._fields == other._fields
            return False

        def __hash__(self):
            return hash(tuple(self._fields))

        class Field:
            _name: str
            _type: Type
            _nullable: bool
            _comment: str

            def __init__(
                self, name: str, field_type: Type, nullable: bool, comment: str
            ) -> None:
                """
                Initializes the Field with the given name, type, nullable flag, and comment.

                Args:
                    name: The name of the field.
                    field_type: The type of the field.
                    nullable: Whether the field is nullable.
                    comment: The comment of the field (optional).
                """
                if name is None:
                    raise ValueError("name cannot be null")
                if type is None:
                    raise ValueError("type cannot be null")
                self._name = name
                self._type = field_type
                self._nullable = nullable
                self._comment = comment

            @classmethod
            def not_null_field(
                cls, name: str, field_type: Type, comment: str = None
            ) -> Types.StructType.Field:
                """
                Args:
                    name: The name of the field.
                    field_type: The type of the field.
                    comment: The comment of the field (optional).
                """
                return cls(name, field_type, False, comment)

            @classmethod
            def nullable_field(
                cls, name: str, field_type: Type, comment: str = None
            ) -> Types.StructType.Field:
                """
                Args:
                    name: The name of the field.
                    field_type: The type of the field.
                    comment: The comment of the field (optional).

                Returns:
                    A nullable Field instance with the given name, field_type, and comment.
                """
                return cls(name, field_type, True, comment)

            def name(self):
                return self._name

            def type(self):
                return self._type

            def nullable(self):
                return self._nullable

            def comment(self):
                return self._comment

            def __eq__(self, other):
                """
                Compares two Field objects for equality.

                Args:
                    other: The other Field object to compare with.

                Returns:
                    True if both Field objects have the same attributes, False otherwise.
                """
                if isinstance(other, Types.StructType.Field):
                    return (
                        self._name == other._name
                        and self._type == other._type
                        and self._nullable == other._nullable
                        and self._comment == other._comment
                    )
                return False

            def __hash__(self):
                return hash((self._name, self._type, self._nullable))

            def simple_string(self) -> str:
                nullable_str = "NULL" if self._nullable else "NOT NULL"
                comment_str = f" COMMENT '{self._comment}'" if self._comment else ""
                return f"{self._name}: {self._type.simple_string()} {nullable_str}{comment_str}"

    class ListType(ComplexType):
        """A list type. Note, this type is not supported in the current version of Gravitino."""

        _element_type: Type
        _element_nullable: bool

        def __init__(self, element_type: Type, element_nullable: bool):
            """
            Create a new ListType with the given element type and the type is nullable.

            Args:
                element_type: The element type of the list.
                element_nullable: Whether the element of the list is nullable.
            """
            if element_type is None:
                raise ValueError("element_type cannot be null")
            self._element_type = element_type
            self._element_nullable = element_nullable

        @classmethod
        def nullable(cls, element_type: Type) -> Types.ListType:
            """
            Create a new ListType with the given element type and the type is nullable.

            Args:
                element_type: The element type of the list.

            Returns:
                A new ListType instance.
            """
            return cls.of(element_type, True)

        @classmethod
        def not_null(cls, element_type: Type) -> Types.ListType:
            """
            Create a new ListType with the given element type.

            Args:
                element_type: The element type of the list.

            Returns:
                A new ListType instance.
            """
            return cls.of(element_type, False)

        @classmethod
        def of(cls, element_type: Type, element_nullable: bool) -> Types.ListType:
            """
            Create a new ListType with the given element type and whether the element is nullable.

            Args:
                element_type: The element type of the list.
                element_nullable: Whether the element of the list is nullable.

            Returns
                A new ListType instance.
            """
            return cls(element_type, element_nullable)

        def element_type(self) -> Type:
            return self._element_type

        def element_nullable(self) -> bool:
            return self._element_nullable

        def name(self) -> Name:
            return Name.LIST

        def simple_string(self) -> str:
            return (
                f"list<{self._element_type.simple_string()}>"
                if self._element_nullable
                else f"list<{self._element_type.simple_string()}, NOT NULL>"
            )

        def __eq__(self, other):
            if not isinstance(other, Types.ListType):
                return (
                    self._element_nullable == other.element_nullable()
                    and self._element_type == other.element_type()
                )
            return False

        def __hash__(self):
            return hash((self._element_type, self._element_nullable))

    class MapType(ComplexType):
        """The map type in Gravitino. Note, this type is not supported in the current version of Gravitino."""

        _key_type: Type
        _value_type: Type
        _value_nullable: bool

        def __init__(self, key_type: Type, value_type: Type, value_nullable: bool):
            """
            Create a new MapType with the given key type, value type and the value is nullable.

            Args:
                key_type: The key type of the map.
                value_type: The value type of the map.
                value_nullable: Whether the value of the map is nullable.
            """
            self._key_type = key_type
            self._value_type = value_type
            self._value_nullable = value_nullable

        @classmethod
        def value_nullable(cls, key_type: Type, value_type: Type) -> Types.MapType:
            """
            Create a new MapType with the given key type, value type, and the value is nullable.

            Args:
                key_type: The key type of the map.
                value_type: The value type of the map.

            Returns:
                A new MapType instance.
            """
            return cls.of(key_type, value_type, True)

        @classmethod
        def value_not_null(cls, key_type: Type, value_type: Type) -> Types.MapType:
            """
            Create a new MapType with the given key type, value type, and the value is not nullable.

            Args:
                key_type: The key type of the map.
                value_type: The value type of the map.

            Returns:
                A new MapType instance.
            """
            return cls.of(key_type, value_type, False)

        @classmethod
        def of(
            cls, key_type: Type, value_type: Type, value_nullable: bool
        ) -> Types.MapType:
            """
            Create a new MapType with the given key type, value type, and whether the value is nullable.

            Args:
                key_type: The key type of the map.
                value_type: The value type of the map.
                value_nullable: Whether the value of the map is nullable.

            Returns:
                A new MapType instance.
            """
            return cls(key_type, value_type, value_nullable)

        def key_type(self) -> Type:
            return self._key_type

        def value_type(self) -> Type:
            return self._value_type

        def is_value_nullable(self) -> bool:
            return self._value_nullable

        def name(self) -> Name:
            return Name.MAP

        def simple_string(self) -> str:
            return f"map<{self._key_type.simple_string()}, {self._value_type.simple_string()}>"

        def __eq__(self, other):
            """
            Compares two MapType objects for equality.

            Args:
                other The other MapType object to compare with.

            Returns:
                True if both MapType objects have the same key type, value type, and nullability, False otherwise.
            """
            if isinstance(other, Types.MapType):
                return (
                    self._value_nullable == other._value_nullable
                    and self._key_type == other._key_type
                    and self._value_type == other._value_type
                )
            return False

        def __hash__(self):
            return hash((self._key_type, self._value_type, self._value_nullable))

    class UnionType(ComplexType):
        """The union type in Gravitino. Note, this type is not supported in the current version of Gravitino."""

        _types: list[Type]

        def __init__(self, types: list[Type]):
            """
            Create a new UnionType with the given types.

            Args:
                types The types of the union.
            """
            self._types = types

        @classmethod
        def of(cls, *types: Type) -> Types.UnionType:
            """
            Create a new UnionType with the given types.

            Args:
                types: The types of the union.

            Returns:
                A new UnionType instance.
            """
            return Types.UnionType(list(types))

        def types(self) -> list:
            return self._types

        def name(self) -> Name:
            return Name.UNION

        def simple_string(self) -> str:
            return f"union<{', '.join(t.simple_string() for t in self._types)}>"

        def __eq__(self, other):
            """
            Compares two UnionType objects for equality.

            Args:
                other The other UnionType object to compare with.

            Returns:
                True if both UnionType objects have the same types, False otherwise.
            """
            if not isinstance(other, Types.UnionType):
                return self._types == other.types()
            return False

        def __hash__(self):
            return hash(tuple(self._types))

    class UnparsedType(Type):
        """Represents a type that is not parsed yet. The parsed type is represented by other types of types."""

        _unparsed_type: str

        def __init__(self, unparsed_type: str):
            """
            Initializes an unparsed_type instance.

            Args:
                unparsed_type: The unparsed type as a string.
            """
            self._unparsed_type = unparsed_type

        @classmethod
        def of(cls, unparsed_type: str) -> Types.UnparsedType:
            """
            Creates a new unparsed_type with the given unparsed type.

            Args:
                unparsed_type The unparsed type.

            Returns
                A new unparsed_type instance.
            """
            return cls(unparsed_type)

        def unparsed_type(self) -> str:
            return self._unparsed_type

        def name(self) -> Name:
            return Name.UNPARSED

        def simple_string(self) -> str:
            return f"unparsed({self._unparsed_type})"

        def __eq__(self, other):
            """
            Compares two unparsed_type objects for equality.

            Args:
                other: The other unparsed_type object to compare with.

            Returns:
                True if both unparsed_type objects have the same unparsed type string, False otherwise.
            """
            if not isinstance(other, Types.UnparsedType):
                return self._unparsed_type == other.unparsed_type()
            return False

        def __hash__(self):
            return hash(self._unparsed_type)

        def __str__(self):
            return self._unparsed_type

    class ExternalType(Type):
        """Represents a type that is defined in an external catalog."""

        _catalog_string: str

        def __init__(self, catalog_string: str):
            """
            Initializes an ExternalType instance.

            Args:
                catalog_string The string representation of this type in the catalog.
            """
            self._catalog_string = catalog_string

        @classmethod
        def of(cls, catalog_string: str) -> Types.ExternalType:
            """
            Creates a new ExternalType with the given catalog string.

            Args:
                catalog_string The string representation of this type in the catalog.

            Returns:
                A new ExternalType instance.
            """
            return cls(catalog_string)

        def catalog_string(self) -> str:
            return self._catalog_string

        def name(self) -> Name:
            return Name.EXTERNAL

        def simple_string(self) -> str:
            return f"external({self._catalog_string})"

        def __eq__(self, other):
            """
            Compares two ExternalType objects for equality.

            Args:
                other: The other ExternalType object to compare with.

            Returns:
                True if both ExternalType objects have the same catalog string, False otherwise.
            """
            if not isinstance(other, Types.ExternalType):
                return False
            return self._catalog_string == other._catalog_string

        def __hash__(self):
            return hash(self._catalog_string)

        def __str__(self):
            return self.simple_string()

    @staticmethod
    def allow_auto_increment(data_type: Type) -> bool:
        """
        Checks if the given data type is allowed to be an auto-increment column.

        Args:
            data_type The data type to check.

        Returns:
            True if the given data type is allowed to be an auto-increment column, False otherwise.
        """
        return isinstance(data_type, (Types.IntegerType, Types.LongType))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from typing import List, Dict

from gravitino.api.schema import Schema
from gravitino.api.schema_change import SchemaChange


class NoSuchSchemaException(Exception):
    """Exception raised if the schema does not exist."""

    pass


class SupportsSchemas(ABC):
    """
    The Catalog interface to support schema operations. If the implemented catalog has schema
    semantics, it should implement this interface.
    """

    @abstractmethod
    def list_schemas(self) -> List[str]:
        """List schemas under the entity.

        If an entity such as a table, view exists, its parent schemas must also exist and must be
        returned by this discovery method. For example, if table a.b.t exists, this method invoked as
        listSchemas(a) must return [b] in the result array

        Raises:
            NoSuchCatalogException: If the catalog does not exist.

        Returns:
            A list of schema names under the namespace.
        """
        pass

    def schema_exists(self, schema_name: str) -> bool:
        """Check if a schema exists.

        If an entity such as a table, view exists, its parent namespaces must also exist. For
        example, if table a.b.t exists, this method invoked as schema_exists(a.b) must return true.

        Args:
            schema_name: The name of the schema.

        Returns:
            True if the schema exists, false otherwise.
        """
        try:
            self.load_schema(schema_name)
            return True
        except NoSuchSchemaException:
            return False

    @abstractmethod
    def create_schema(
        self, schema_name: str, comment: str, properties: Dict[str, str]
    ) -> Schema:
        """Create a schema in the catalog.

        Args:
            schema_name: The name of the schema.
            comment: The comment of the schema.
            properties: The properties of the schema.

        Raises:
            NoSuchCatalogException: If the catalog does not exist.
            SchemaAlreadyExistsException: If the schema already exists.

        Returns:
            The created schema.
        """
        pass

    @abstractmethod
    def load_schema(self, schema_name: str) -> Schema:
        """Load metadata properties for a schema.

        Args:
            schema_name: The name of the schema.

        Raises:
            NoSuchSchemaException: If the schema does not exist (optional).

        Returns:
            A schema.
        """
        pass

    @abstractmethod
    def alter_schema(self, schema_name: str, *changes: SchemaChange) -> Schema:
        """Apply the metadata change to a schema in the catalog.

        Args:
            schema_name: The name of the schema.
            changes: The metadata changes to apply.

        Raises:
            NoSuchSchemaException: If the schema does not exist.

        Returns:
            The altered schema.
        """
        pass

    @abstractmethod
    def drop_schema(self, schema_name: str, cascade: bool) -> bool:
        """Drop a schema from the catalog. If cascade option is true, recursively
        drop all objects within the schema.

        Args:
            schema_name: The name of the schema.
            cascade: If true, recursively drop all objects within the schema.

        Returns:
            True if the schema exists and is dropped successfully, false otherwise.

        Raises:
            NonEmptySchemaException: If the schema is not empty and cascade is false.
        """
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import re
from dataclasses import dataclass
from enum import Enum

from gravitino.dto.version_dto import VersionDTO
from gravitino.exceptions.base import BadRequestException, GravitinoRuntimeException


class Version(Enum):
    MAJOR = "MAJOR"
    MINOR = "MINOR"
    PATCH = "PATCH"


VERSION_PATTERN: str = (
    rf"(?P<{Version.MAJOR.value}>\d+)\.(?P<{Version.MINOR.value}>\d+)\.(?P<{Version.PATCH.value}>\d+)+?"
)


@dataclass
class GravitinoVersion(VersionDTO):
    """Gravitino version information."""

    major: int
    minor: int
    patch: int

    def __init__(self, versionDTO):
        super().__init__(
            versionDTO.version(), versionDTO.compile_date(), versionDTO.git_commit()
        )

        m = re.match(VERSION_PATTERN, self.version())

        if m is None:
            raise BadRequestException(f"Invalid version string: {self.version()}")

        self.major = int(m.group(Version.MAJOR.value))
        self.minor = int(m.group(Version.MINOR.value))
        self.patch = int(m.group(Version.PATCH.value))

    def __gt__(self, other) -> bool:
        if not isinstance(other, GravitinoVersion):
            raise GravitinoRuntimeException(
                f"{GravitinoVersion.__name__} can't compare with {other.__class__.__name__}"
            )
        if self.major > other.major:
            return True
        if self.minor > other.minor:
            return True
        if self.patch > other.patch:
            return True
        return False

    def __eq__(self, other) -> bool:
        if not isinstance(other, GravitinoVersion):
            raise GravitinoRuntimeException(
                f"{GravitinoVersion.__name__} can't compare with {other.__class__.__name__}"
            )
        if self.major != other.major:
            return False
        if self.minor != other.minor:
            return False
        if self.patch != other.patch:
            return False
        return True



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from abc import ABC, abstractmethod
from typing import List

from gravitino.api.expressions.expression import Expression
from gravitino.api.expressions.sorts.null_ordering import NullOrdering
from gravitino.api.expressions.sorts.sort_direction import SortDirection


class SortOrder(Expression, ABC):
    """Represents a sort order in the public expression API."""

    @abstractmethod
    def expression(self) -> Expression:
        """Returns the sort expression."""
        pass

    @abstractmethod
    def direction(self) -> SortDirection:
        """Returns the sort direction."""
        pass

    @abstractmethod
    def null_ordering(self) -> NullOrdering:
        """Returns the null ordering."""
        pass

    def children(self) -> List[Expression]:
        """Returns the children expressions of this sort order."""
        return [self.expression()]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Dict
from dataclasses import dataclass, field

from dataclasses_json import config

from .audit_dto import AuditDTO
from ..api.catalog import Catalog


@dataclass
class CatalogDTO(Catalog):
    """Data transfer object representing catalog information."""

    _name: str = field(metadata=config(field_name="name"))
    _type: Catalog.Type = field(
        metadata=config(
            field_name="type",
            encoder=Catalog.Type.type_serialize,
            decoder=Catalog.Type.type_deserialize,
        )
    )
    _provider: str = field(metadata=config(field_name="provider"))
    _comment: str = field(metadata=config(field_name="comment"))
    _properties: Dict[str, str] = field(metadata=config(field_name="properties"))
    _audit: AuditDTO = field(default=None, metadata=config(field_name="audit"))

    def builder(
        self,
        name: str = None,
        catalog_type: Catalog.Type = Catalog.Type.UNSUPPORTED,
        provider: str = None,
        comment: str = None,
        properties: Dict[str, str] = None,
        audit: AuditDTO = None,
    ):
        self._name = name
        self._type = catalog_type
        self._provider = provider
        self._comment = comment
        self._properties = properties
        self._audit = audit

    def name(self) -> str:
        return self._name

    def type(self) -> Catalog.Type:
        return self._type

    def provider(self) -> str:
        return self._provider

    def comment(self) -> str:
        return self._comment

    def properties(self) -> Dict[str, str]:
        return self._properties

    def audit_info(self) -> AuditDTO:
        return self._audit



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC
from typing import Dict

from gravitino.api.credential.credential import Credential
from gravitino.utils.precondition import Precondition


class AzureAccountKeyCredential(Credential, ABC):
    """Represents Azure account key credential."""

    AZURE_ACCOUNT_KEY_CREDENTIAL_TYPE: str = "azure-account-key"
    _STORAGE_ACCOUNT_NAME: str = "azure-storage-account-name"
    _STORAGE_ACCOUNT_KEY: str = "azure-storage-account-key"

    def __init__(self, credential_info: Dict[str, str], expire_time_in_ms: int):
        self._account_name = credential_info.get(self._STORAGE_ACCOUNT_NAME, None)
        self._account_key = credential_info.get(self._STORAGE_ACCOUNT_KEY, None)
        Precondition.check_string_not_empty(
            self._account_name, "The Azure account name should not be empty"
        )
        Precondition.check_string_not_empty(
            self._account_key, "The Azure account key should not be empty"
        )
        Precondition.check_argument(
            expire_time_in_ms == 0,
            "The expiration time of Azure account key credential should be 0",
        )

    def credential_type(self) -> str:
        """Returns the type of the credential.

        Returns:
            The type of the credential.
        """
        return self.AZURE_ACCOUNT_KEY_CREDENTIAL_TYPE

    def expire_time_in_ms(self) -> int:
        """Returns the expiration time of the credential in milliseconds since
        the epoch, 0 means it will never expire.

        Returns:
            The expiration time of the credential.
        """
        return 0

    def credential_info(self) -> Dict[str, str]:
        """The credential information.

        Returns:
            The credential information.
        """
        return {
            self._STORAGE_ACCOUNT_NAME: self._account_name,
            self._STORAGE_ACCOUNT_KEY: self._account_key,
        }

    def account_name(self) -> str:
        """The Azure account name.

        Returns:
            The Azure account name.
        """
        return self._account_name

    def account_key(self) -> str:
        """The Azure account key.

        Returns:
            The Azure account key.
        """
        return self._account_key



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from dataclasses import dataclass, field
from typing import List

from dataclasses_json import config

from gravitino.dto.responses.base_response import BaseResponse
from gravitino.exceptions.base import IllegalArgumentException


@dataclass
class ModelVersionListResponse(BaseResponse):
    """Represents a response for a list of model versions."""

    _versions: List[int] = field(metadata=config(field_name="versions"))

    def versions(self) -> List[int]:
        return self._versions

    def validate(self):
        """Validates the response data.

        Raises:
            IllegalArgumentException if versions are not set.
        """
        super().validate()

        if self._versions is None:
            raise IllegalArgumentException("versions must not be null")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
from typing import List, Dict

from gravitino.client.gravitino_client_base import GravitinoClientBase
from gravitino.client.gravitino_metalake import GravitinoMetalake
from gravitino.client.dto_converters import DTOConverters
from gravitino.dto.requests.metalake_create_request import MetalakeCreateRequest
from gravitino.dto.requests.metalake_set_request import MetalakeSetRequest
from gravitino.dto.requests.metalake_updates_request import MetalakeUpdatesRequest
from gravitino.dto.responses.drop_response import DropResponse
from gravitino.dto.responses.metalake_list_response import MetalakeListResponse
from gravitino.dto.responses.metalake_response import MetalakeResponse
from gravitino.api.metalake_change import MetalakeChange
from gravitino.exceptions.handlers.metalake_error_handler import METALAKE_ERROR_HANDLER
from gravitino.rest.rest_utils import encode_string

logger = logging.getLogger(__name__)


class GravitinoAdminClient(GravitinoClientBase):
    """
    Gravitino Client for the administrator to interact with the Gravitino API.
    It allows the client to list, load, create, and alter Metalakes.
    Normal users should use {@link GravitinoClient} to connect with the Gravitino server.
    """

    def list_metalakes(self) -> List[GravitinoMetalake]:
        """Retrieves a list of Metalakes from the Gravitino API.

        Returns:
            An array of GravitinoMetalake objects representing the Metalakes.
        """
        resp = self._rest_client.get(
            self.API_METALAKES_LIST_PATH, error_handler=METALAKE_ERROR_HANDLER
        )
        metalake_list_resp = MetalakeListResponse.from_json(
            resp.body, infer_missing=True
        )
        metalake_list_resp.validate()

        return [
            GravitinoMetalake(o, self._rest_client)
            for o in metalake_list_resp.metalakes()
        ]

    def create_metalake(
        self, name: str, comment: str, properties: Dict[str, str]
    ) -> GravitinoMetalake:
        """Creates a new Metalake using the Gravitino API.

        Args:
            name: The name of the new Metalake.
            comment: The comment for the new Metalake.
            properties: The properties of the new Metalake.

        Returns:
            A GravitinoMetalake instance representing the newly created Metalake.
            TODO: @throws MetalakeAlreadyExistsException If a Metalake with the specified identifier already exists.
        """
        req = MetalakeCreateRequest(name, comment, properties)
        req.validate()

        resp = self._rest_client.post(
            self.API_METALAKES_LIST_PATH, req, error_handler=METALAKE_ERROR_HANDLER
        )
        metalake_response = MetalakeResponse.from_json(resp.body, infer_missing=True)
        metalake_response.validate()
        metalake = metalake_response.metalake()

        return GravitinoMetalake(metalake, self._rest_client)

    def alter_metalake(self, name: str, *changes: MetalakeChange) -> GravitinoMetalake:
        """Alters a specific Metalake using the Gravitino API.

        Args:
            name: The name of the Metalake to be altered.
            changes: The changes to be applied to the Metalake.

        Returns:
             A GravitinoMetalake instance representing the updated Metalake.
        TODO: @throws NoSuchMetalakeException If the specified Metalake does not exist.
        TODO: @throws IllegalArgumentException If the provided changes are invalid or not applicable.
        """

        reqs = [DTOConverters.to_metalake_update_request(change) for change in changes]
        updates_request = MetalakeUpdatesRequest(reqs)
        updates_request.validate()

        resp = self._rest_client.put(
            self.API_METALAKES_IDENTIFIER_PATH + encode_string(name),
            updates_request,
            error_handler=METALAKE_ERROR_HANDLER,
        )
        metalake_response = MetalakeResponse.from_json(resp.body, infer_missing=True)
        metalake_response.validate()
        metalake = metalake_response.metalake()

        return GravitinoMetalake(metalake, self._rest_client)

    def drop_metalake(self, name: str, force: bool = False) -> bool:
        """Drops a specific Metalake using the Gravitino API.

        Args:
            name: The name of the Metalake to be dropped.
            force: Whether to force the drop operation.

        Returns:
            True if the Metalake was successfully dropped, false if the Metalake does not exist.
        """

        params = {"force": str(force)}
        resp = self._rest_client.delete(
            self.API_METALAKES_IDENTIFIER_PATH + encode_string(name),
            params=params,
            error_handler=METALAKE_ERROR_HANDLER,
        )
        drop_response = DropResponse.from_json(resp.body, infer_missing=True)

        return drop_response.dropped()

    def enable_metalake(self, name: str):
        """Enable the metalake with specified name. If the metalake is already in use, this method does nothing.

        Args:
            name: the name of the metalake.

        Raises:
            NoSuchMetalakeException if the metalake with specified name does not exist.
        """

        metalake_enable_request = MetalakeSetRequest(in_use=True)
        metalake_enable_request.validate()

        url = self.API_METALAKES_IDENTIFIER_PATH + encode_string(name)
        self._rest_client.patch(
            url, json=metalake_enable_request, error_handler=METALAKE_ERROR_HANDLER
        )

    def disable_metalake(self, name: str):
        """Disable the metalake with specified name. If the metalake is already disabled, does nothing.

        Args:
            name: the name of the metalake.

        Raises:
            NoSuchMetalakeException if the metalake with specified name does not exist.
        """

        metalake_disable_request = MetalakeSetRequest(in_use=False)
        metalake_disable_request.validate()

        url = self.API_METALAKES_IDENTIFIER_PATH + encode_string(name)
        self._rest_client.patch(
            url, json=metalake_disable_request, error_handler=METALAKE_ERROR_HANDLER
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from typing import List

from dataclasses_json import config

from gravitino.dto.requests.metalake_update_request import MetalakeUpdateRequest
from gravitino.rest.rest_message import RESTRequest


@dataclass
class MetalakeUpdatesRequest(RESTRequest):
    """Represents a request containing multiple Metalake updates."""

    _updates: List[MetalakeUpdateRequest] = field(metadata=config(field_name="updates"))

    def __init__(self, updates: List[MetalakeUpdateRequest]):
        """Constructor for MetalakeUpdatesRequest.

        Args:
            updates: The list of Metalake update requests.
        """
        self._updates = updates

    def validate(self):
        """Validates each request in the list.

        Raises:
            IllegalArgumentException if validation of any request fails.
        """
        for update in self._updates:
            update.validate()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import List, ClassVar
from gravitino.exceptions.base import IllegalNamespaceException


class Namespace:
    """A namespace is a sequence of levels separated by dots. It's used to identify a metalake, a
    catalog or a schema. For example, "metalake1", "metalake1.catalog1" and
    "metalake1.catalog1.schema1" are all valid namespaces.
    """

    _DOT: ClassVar[str] = "."

    _levels: List[str]

    def __init__(self, levels: List[str]):
        self._levels = levels

    def to_json(self):
        return self._levels

    @classmethod
    def from_json(cls, levels):
        if levels is None or not isinstance(levels, list):
            raise IllegalNamespaceException(
                f"Cannot parse namespace from invalid JSON: {levels}"
            )
        return cls(levels)

    @staticmethod
    def empty() -> "Namespace":
        """Get an empty namespace.

        Returns:
            An empty namespace
        """
        return Namespace([])

    @staticmethod
    def of(*levels: str) -> "Namespace":
        """Create a namespace with the given levels.

        Args:
            levels The levels of the namespace

        Returns:
            A namespace with the given levels
        """
        Namespace.check(
            levels is not None, "Cannot create a namespace with null levels"
        )
        if len(levels) == 0:
            return Namespace.empty()

        for level in levels:
            Namespace.check(
                level is not None and level != "",
                "Cannot create a namespace with null or empty level",
            )

        return Namespace(list(levels))

    def levels(self) -> List[str]:
        """Get the levels of the namespace.

        Returns:
            The levels of the namespace
        """
        return self._levels

    def level(self, pos: int) -> str:
        """Get the level at the given position.

        Args:
            pos: The position of the level

        Returns:
            The level at the given position
        """
        if pos < 0 or pos >= len(self._levels):
            raise ValueError("Invalid level position")
        return self._levels[pos]

    def length(self) -> int:
        """Get the length of the namespace.

        Returns:
            The length of the namespace.
        """
        return len(self._levels)

    def is_empty(self) -> bool:
        """Check if the namespace is empty.

        Returns:
            True if the namespace is empty, false otherwise.
        """
        return len(self._levels) == 0

    def __eq__(self, other: "Namespace") -> bool:
        if not isinstance(other, Namespace):
            return False
        return self._levels == other._levels

    def __hash__(self) -> int:
        return hash(tuple(self._levels))

    def __str__(self) -> str:
        return self._DOT.join(self._levels)

    @staticmethod
    def check(expression: bool, message: str, *args) -> None:
        """Check the given condition is true. Throw an IllegalNamespaceException if it's not.

        Args:
            expression: The expression to check.
            message: The message to throw.
            args: The arguments to the message.
        """
        if not expression:
            raise IllegalNamespaceException(message.format(*args))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field

from dataclasses_json import config


class MetalakeChange:
    """A metalake change is a change to a metalake. It can be used to rename a metalake, update the
    comment of a metalake, set a property and value pair for a metalake, or remove a property from a
    metalake.
    """

    @staticmethod
    def rename(new_name: str) -> "MetalakeChange.RenameMetalake":
        """Creates a new metalake change to rename the metalake.

        Args:
            new_name: The New name of the metalake.

        Returns:
            The metalake change.
        """
        return MetalakeChange.RenameMetalake(new_name)

    @staticmethod
    def update_comment(new_comment: str) -> "MetalakeChange.UpdateMetalakeComment":
        """Creates a new metalake change to update the metalake comment.

        Args:
            new_comment: The new comment of the metalake.

        Returns:
            The metalake change.
        """
        return MetalakeChange.UpdateMetalakeComment(new_comment)

    @staticmethod
    def set_property(metalake_property: str, value: str) -> "SetProperty":
        """Creates a new metalake change to set a property and value pair for the metalake.

        Args:
            metalake_property: The property name to set.
            value: The value to set the property to.

        Returns:
             The metalake change.
        """
        return MetalakeChange.SetProperty(metalake_property, value)

    @staticmethod
    def remove_property(metalake_property: str) -> "RemoveProperty":
        """Creates a new metalake change to remove a property from the metalake.

        Args:
            metalake_property: The property name to remove.

        Returns:
            The metalake change.
        """
        return MetalakeChange.RemoveProperty(metalake_property)

    @dataclass(frozen=True)
    class RenameMetalake:
        """A metalake change to rename the metalake."""

        _new_name: str = field(metadata=config(field_name="new_name"))

        def new_name(self) -> str:
            return self._new_name

        def __str__(self):
            return f"RENAMEMETALAKE {self._new_name}"

    @dataclass(frozen=True)
    class UpdateMetalakeComment:
        """A metalake change to update the metalake comment"""

        _new_comment: str = field(metadata=config(field_name="new_comment"))

        def new_comment(self) -> str:
            return self._new_comment

        def __str__(self):
            return f"UPDATEMETALAKECOMMENT {self._new_comment}"

    @dataclass(frozen=True)
    class SetProperty:
        """A metalake change to set a property and value pair for the metalake"""

        _property: str = field(metadata=config(field_name="property"))
        _value: str = field(metadata=config(field_name="value"))

        def property(self) -> str:
            return self._property

        def value(self) -> str:
            return self._value

        def __str__(self):
            return f"SETPROPERTY {self._property} {self._value}"

    @dataclass(frozen=True)
    class RemoveProperty:
        """A metalake change to remove a property from the metalake"""

        _property: str

        def property(self) -> str:
            return self._property

        def __str__(self):
            return f"REMOVEPROPERTY {self.property}"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import abstractmethod
from enum import Enum
from typing import Optional, Dict

from gravitino.api.auditable import Auditable


class Fileset(Auditable):
    """An interface representing a fileset under a schema Namespace. A fileset is a virtual
    concept of the file or directory that is managed by Gravitino. Users can create a fileset object
    to manage the non-tabular data on the FS-like storage. The typical use case is to manage the
    training data for AI workloads. The major difference compare to the relational table is that the
    fileset is schema-free, the main property of the fileset is the storage location of the
    underlying data.

    Fileset defines the basic properties of a fileset object. A catalog implementation
    with FilesetCatalog should implement this interface.
    """

    class Type(Enum):
        """An enum representing the type of the fileset object."""

        MANAGED = "managed"
        """Fileset is managed by Gravitino. 
        When specified, the data will be deleted when the fileset object is deleted"""

        EXTERNAL = "external"
        """Fileset is not managed by Gravitino. 
        When specified, the data will not be deleted when the fileset object is deleted"""

    @abstractmethod
    def name(self) -> str:
        """
        Returns:
            Name of the fileset object.
        """
        pass

    @abstractmethod
    def comment(self) -> Optional[str]:
        """
        Returns:
            The comment of the fileset object. Null is returned if no comment is set.
        """
        pass

    @abstractmethod
    def type(self) -> Type:
        """
        @Returns:
            The type of the fileset object.
        """
        pass

    @abstractmethod
    def storage_location(self) -> str:
        """Get the storage location of the file or directory path that is managed by this fileset object.

        The returned storageLocation can either be the one specified when creating the fileset
        object, or the one specified in the catalog / schema level if the fileset object is created
        under this catalog / schema.

        For managed fileset, the storageLocation can be:

        1) The one specified when creating the fileset object.

        2) When catalog property "location" is specified but schema property "location" is not
        specified, then the storageLocation will be "{catalog location}/schemaName/filesetName".

        3) When catalog property "location" is not specified but schema property "location" is
        specified, then the storageLocation will be "{schema location}/filesetName".

        4) When both catalog property "location" and schema property "location" are specified, then
        the storageLocation will be "{schema location}/filesetName".

        5) When both catalog property "location" and schema property "location" are not specified,
        and storageLocation specified when creating the fileset object is null, this situation is
        illegal.

        For external fileset, the storageLocation can be:

        1) The one specified when creating the fileset object.

        Returns:
             The storage location of the fileset object.
        """
        pass

    @abstractmethod
    def properties(self) -> Dict[str, str]:
        """
        Returns:
            The properties of the fileset object. Empty map is returned if no properties are set.
        """
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
from typing import List, Dict

from gravitino.api.catalog import Catalog
from gravitino.api.catalog_change import CatalogChange
from gravitino.client.dto_converters import DTOConverters
from gravitino.dto.metalake_dto import MetalakeDTO
from gravitino.dto.requests.catalog_create_request import CatalogCreateRequest
from gravitino.dto.requests.catalog_set_request import CatalogSetRequest
from gravitino.dto.requests.catalog_updates_request import CatalogUpdatesRequest
from gravitino.dto.responses.catalog_list_response import CatalogListResponse
from gravitino.dto.responses.catalog_response import CatalogResponse
from gravitino.dto.responses.drop_response import DropResponse
from gravitino.dto.responses.entity_list_response import EntityListResponse
from gravitino.exceptions.handlers.catalog_error_handler import CATALOG_ERROR_HANDLER
from gravitino.rest.rest_utils import encode_string
from gravitino.utils import HTTPClient


logger = logging.getLogger(__name__)


class GravitinoMetalake(MetalakeDTO):
    """
    Gravitino Metalake is the top-level metadata repository for users. It contains a list of catalogs
    as sub-level metadata collections. With GravitinoMetalake, users can list, create, load,
    alter and drop a catalog with specified identifier.
    """

    rest_client: HTTPClient

    API_METALAKES_CATALOGS_PATH = "api/metalakes/{}/catalogs/{}"

    def __init__(self, metalake: MetalakeDTO = None, client: HTTPClient = None):
        super().__init__(
            _name=metalake.name(),
            _comment=metalake.comment(),
            _properties=metalake.properties(),
            _audit=metalake.audit_info(),
        )
        self.rest_client = client

    def list_catalogs(self) -> List[str]:
        """List all the catalogs under this metalake.

        Raises:
            NoSuchMetalakeException if the metalake with specified namespace does not exist.

        Returns:
            A list of the catalog names under this metalake.
        """
        url = f"api/metalakes/{self.name()}/catalogs"
        response = self.rest_client.get(url, error_handler=CATALOG_ERROR_HANDLER)
        entity_list = EntityListResponse.from_json(response.body, infer_missing=True)
        entity_list.validate()
        return [identifier.name() for identifier in entity_list.identifiers()]

    def list_catalogs_info(self) -> List[Catalog]:
        """List all the catalogs with their information under this metalake.

        Raises:
            NoSuchMetalakeException if the metalake with specified namespace does not exist.

        Returns:
            A list of Catalog under the specified namespace.
        """
        params = {"details": "true"}
        url = f"api/metalakes/{self.name()}/catalogs"
        response = self.rest_client.get(
            url, params=params, error_handler=CATALOG_ERROR_HANDLER
        )
        catalog_list = CatalogListResponse.from_json(response.body, infer_missing=True)

        return [
            DTOConverters.to_catalog(self.name(), catalog, self.rest_client)
            for catalog in catalog_list.catalogs()
        ]

    def load_catalog(self, name: str) -> Catalog:
        """Load the catalog with specified name.

        Args:
            name: The name of the catalog to load.

        Raises:
            NoSuchCatalogException if the catalog with specified name does not exist.

        Returns:
            The Catalog with specified name.
        """
        url = self.API_METALAKES_CATALOGS_PATH.format(
            encode_string(self.name()), encode_string(name)
        )
        response = self.rest_client.get(url, error_handler=CATALOG_ERROR_HANDLER)
        catalog_resp = CatalogResponse.from_json(response.body, infer_missing=True)

        return DTOConverters.to_catalog(
            self.name(), catalog_resp.catalog(), self.rest_client
        )

    def create_catalog(
        self,
        name: str,
        catalog_type: Catalog.Type,
        provider: str,
        comment: str,
        properties: Dict[str, str],
    ) -> Catalog:
        """Create a new catalog with specified name, catalog type, comment and properties.

        Args:
            name: The name of the catalog.
            catalog_type: The type of the catalog.
            provider: The provider of the catalog. This parameter can be None if the catalog
            provides a managed implementation. Currently, only the model catalog supports None
            provider. For the details, please refer to the Catalog.Type.
            comment: The comment of the catalog.
            properties: The properties of the catalog.

        Raises:
            NoSuchMetalakeException if the metalake does not exist.
            CatalogAlreadyExistsException if the catalog with specified name already exists.

        Returns:
            The created Catalog.
        """

        catalog_create_request = CatalogCreateRequest(
            name=name,
            catalog_type=catalog_type,
            provider=provider,
            comment=comment,
            properties=properties,
        )
        catalog_create_request.validate()

        url = f"api/metalakes/{encode_string(self.name())}/catalogs"
        response = self.rest_client.post(
            url, json=catalog_create_request, error_handler=CATALOG_ERROR_HANDLER
        )
        catalog_resp = CatalogResponse.from_json(response.body, infer_missing=True)

        return DTOConverters.to_catalog(
            self.name(), catalog_resp.catalog(), self.rest_client
        )

    def alter_catalog(self, name: str, *changes: CatalogChange) -> Catalog:
        """Alter the catalog with specified name by applying the changes.

        Args:
            name: the name of the catalog.
            changes: the changes to apply to the catalog.

        Raises:
            NoSuchCatalogException if the catalog with specified name does not exist.
            IllegalArgumentException if the changes are invalid.

        Returns:
            the altered Catalog.
        """

        reqs = [DTOConverters.to_catalog_update_request(change) for change in changes]
        updates_request = CatalogUpdatesRequest(reqs)
        updates_request.validate()

        url = self.API_METALAKES_CATALOGS_PATH.format(
            encode_string(self.name()), encode_string(name)
        )
        response = self.rest_client.put(
            url, json=updates_request, error_handler=CATALOG_ERROR_HANDLER
        )
        catalog_response = CatalogResponse.from_json(response.body, infer_missing=True)
        catalog_response.validate()

        return DTOConverters.to_catalog(
            self.name(), catalog_response.catalog(), self.rest_client
        )

    def drop_catalog(self, name: str, force: bool = False) -> bool:
        """Drop the catalog with specified name.

        Args:
            name: the name of the catalog.
            force: whether to force drop the catalog.

        Returns:
            true if the catalog is dropped successfully, false if the catalog does not exist.
        """
        params = {"force": str(force)}
        url = self.API_METALAKES_CATALOGS_PATH.format(
            encode_string(self.name()), encode_string(name)
        )
        response = self.rest_client.delete(
            url, params=params, error_handler=CATALOG_ERROR_HANDLER
        )

        drop_response = DropResponse.from_json(response.body, infer_missing=True)
        drop_response.validate()

        return drop_response.dropped()

    def enable_catalog(self, name: str):
        """Enable the catalog with specified name. If the catalog is already in use, this method does nothing.

        Args:
            name: the name of the catalog.

        Raises:
            NoSuchCatalogException if the catalog with specified name does not exist.
        """

        catalog_enable_request = CatalogSetRequest(in_use=True)
        catalog_enable_request.validate()

        url = self.API_METALAKES_CATALOGS_PATH.format(
            encode_string(self.name()), encode_string(name)
        )
        self.rest_client.patch(
            url, json=catalog_enable_request, error_handler=CATALOG_ERROR_HANDLER
        )

    def disable_catalog(self, name: str):
        """Disable the catalog with specified name. If the catalog is already disabled, this method does nothing.

        Args:
            name: the name of the catalog.

        Raises:
            NoSuchCatalogException if the catalog with specified name does not exist.
        """

        catalog_disable_request = CatalogSetRequest(in_use=False)
        catalog_disable_request.validate()

        url = self.API_METALAKES_CATALOGS_PATH.format(
            encode_string(self.name()), encode_string(name)
        )
        self.rest_client.patch(
            url, json=catalog_disable_request, error_handler=CATALOG_ERROR_HANDLER
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import decimal
from typing import TypeVar
from datetime import date, time, datetime

from gravitino.api.expressions.literals.literal import Literal
from gravitino.api.types.type import Type
from gravitino.api.types.types import Types

T = TypeVar("T")


class LiteralImpl(Literal[T]):
    """Creates a literal with the given type value."""

    _value: T
    _data_type: Type

    def __init__(self, value: T, data_type: Type):
        self._value = value
        self._data_type = data_type

    def value(self) -> T:
        return self._value

    def data_type(self) -> Type:
        return self._data_type

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, LiteralImpl):
            return False
        return (self._value == other._value) and (self._data_type == other._data_type)

    def __hash__(self):
        return hash((self._value, self._data_type))

    def __str__(self):
        return f"LiteralImpl(value={self._value}, data_type={self._data_type})"


class Literals:
    """The helper class to create literals to pass into Apache Gravitino."""

    NULL = LiteralImpl(None, Types.NullType.get())

    @staticmethod
    def of(value: T, data_type: Type) -> Literal[T]:
        return LiteralImpl(value, data_type)

    @staticmethod
    def boolean_literal(value: bool) -> LiteralImpl[bool]:
        return LiteralImpl(value, Types.BooleanType.get())

    @staticmethod
    def byte_literal(value: str) -> LiteralImpl[str]:
        return LiteralImpl(value, Types.ByteType.get())

    @staticmethod
    def unsigned_byte_literal(value: str) -> LiteralImpl[str]:
        return LiteralImpl(value, Types.ByteType.unsigned())

    @staticmethod
    def short_literal(value: int) -> LiteralImpl[int]:
        return LiteralImpl(value, Types.ShortType.get())

    @staticmethod
    def unsigned_short_literal(value: int) -> LiteralImpl[int]:
        return LiteralImpl(value, Types.ShortType.unsigned())

    @staticmethod
    def integer_literal(value: int) -> LiteralImpl[int]:
        return LiteralImpl(value, Types.IntegerType.get())

    @staticmethod
    def unsigned_integer_literal(value: int) -> LiteralImpl[int]:
        return LiteralImpl(value, Types.IntegerType.unsigned())

    @staticmethod
    def long_literal(value: int) -> LiteralImpl[int]:
        return LiteralImpl(value, Types.LongType.get())

    @staticmethod
    def unsigned_long_literal(value: int) -> LiteralImpl[int]:
        return LiteralImpl(value, Types.LongType.unsigned())

    @staticmethod
    def float_literal(value: float) -> LiteralImpl[float]:
        return LiteralImpl(value, Types.FloatType.get())

    @staticmethod
    def double_literal(value: float) -> LiteralImpl[float]:
        return LiteralImpl(value, Types.DoubleType.get())

    @staticmethod
    def decimal_literal(value: decimal.Decimal) -> LiteralImpl[decimal.Decimal]:
        precision: int = len(value.as_tuple().digits)
        scale: int = -value.as_tuple().exponent
        return LiteralImpl(value, Types.DecimalType.of(max(precision, scale), scale))

    @staticmethod
    def date_literal(value: date) -> Literal[date]:
        return LiteralImpl(value, Types.DateType.get())

    @staticmethod
    def time_literal(value: time) -> Literal[time]:
        return Literals.of(value, Types.TimeType.get())

    @staticmethod
    def timestamp_literal(value: datetime) -> Literal[datetime]:
        return Literals.of(value, Types.TimestampType.without_time_zone())

    @staticmethod
    def timestamp_literal_from_string(value: str) -> Literal[datetime]:
        return Literals.timestamp_literal(datetime.fromisoformat(value))

    @staticmethod
    def string_literal(value: str) -> Literal[str]:
        return LiteralImpl(value, Types.StringType.get())

    @staticmethod
    def varchar_literal(length: int, value: str) -> Literal[str]:
        return LiteralImpl(value, Types.VarCharType.of(length))
